import base64
import binascii
import io
import json
import os
import re
import time
import warnings
from collections.abc import Collection, Mapping
from typing import Any, Dict, Tuple, Type, Union

import numpy as np
import pandas as pd
import polars as pl
import requests

try:
    import geopandas as gpd
except ImportError:
    gpd = None

MSG = (
    "Could not import geopandas.  Must install nuclei[geo] "
    " in order to use validate method"
)

DF = "DataFrame"
DF_PARQUET = "DataFrame.parquet"
PDS = "Series"
NUMPY = "numpy"
GDF = "GeoDataFrame"
GPDS = "GeoSeries"
POLARS_DF = "POLARS_DataFrame"
PANDAS_DF = "PANDAS_DataFrame"
PANDAS_DF_PARQUET = "PANDAS_DataFrame.parquet"
PANDAS_PDS = "PANDAS_Series"


def buffer_to_base64(buf: io.BytesIO) -> str:
    buf.seek(0)
    return base64.b64encode(buf.read()).decode("utf-8")


def base64_to_buffer(blob: bytes) -> io.BytesIO:
    blob = base64.b64decode(blob)
    buf = io.BytesIO()
    buf.write(blob)
    return buf


def deserialize_polars_json(message: dict) -> Union[pd.DataFrame, pd.Series]:
    """
    Only needed for backwards compatibility.
    """
    return pl.read_json(message["body"])


def serialize_polars_json(obj: pl.DataFrame) -> dict:
    return {"nuclei": {"type": POLARS_DF}, "body": obj.to_json()}


def deserialize_pandas_json(message: dict) -> Union[pd.DataFrame, pd.Series]:
    """
    Only needed for backwards compatibility.
    """
    typ = "frame"
    if message["nuclei"]["type"] == PDS or message["nuclei"]["type"] == PANDAS_PDS:
        typ = "series"

    return pd.read_json(message["body"], typ=typ).sort_index()


def serialize_pandas_json(obj: Union[pd.DataFrame, pd.Series]) -> dict:
    if isinstance(obj, pd.DataFrame):
        return {"nuclei": {"type": PANDAS_DF}, "body": obj.to_json()}
    else:
        return {"nuclei": {"type": PANDAS_PDS}, "body": obj.to_json()}


def serialize_pandas_parquet(obj: Union[pd.DataFrame, pd.Series]) -> dict:
    buf = io.BytesIO()
    if isinstance(obj, pd.Series):
        obj = obj.to_frame(obj.name)
    try:
        obj.to_parquet(buf, index=True)
    except Exception:
        return {"nuclei": {"type": PANDAS_DF}, "body": obj.to_json()}

    return {"nuclei": {"type": PANDAS_DF_PARQUET}, "body": buffer_to_base64(buf)}


def deserialize_pandas_parquet(message: dict) -> pd.DataFrame:
    buf = base64_to_buffer(message["body"])
    return pd.read_parquet(buf)


def deserialize_geopandas_json(
    message: dict,
) -> Union[Type["gpd.GeoDataFrame"], Type["gpd.GeoSeries"], Any]:
    """
    Only needed for backwards compatibility.
    """
    if gpd is None:
        raise ImportError(MSG)

    obj = gpd.read_file(message["body"]).reset_index(drop=True)
    obj = obj.drop(columns="id")

    if len(obj.columns) == 1 and obj.columns[0] == "geometry":
        obj = obj.geometry

    return obj


def serialize_geopandas_json(
    obj: Union[Type["gpd.GeoDataFrame"], Type["gpd.GeoSeries"]]
) -> dict:
    if gpd is None:
        raise ImportError(MSG)

    if isinstance(obj, gpd.GeoDataFrame):
        return {"nuclei": {"type": GDF}, "body": obj.to_json()}
    else:
        return {"nuclei": {"type": GPDS}, "body": obj.to_json()}


def serialize_numpy(array: np.ndarray) -> dict:
    buf = io.BytesIO()
    np.save(buf, array, allow_pickle=False)
    return {"nuclei": {"type": NUMPY}, "body": buffer_to_base64(buf)}


def deserialize_numpy(message: dict) -> np.ndarray:
    buf = base64_to_buffer(message["body"])
    buf.seek(0)
    return np.load(buf, allow_pickle=False)


def python_types_to_message(
    schema: Union[
        dict,
        pd.Series,
        pd.DataFrame,
        pl.DataFrame,
        Type["gpd.GeoSeries"],
        Type["gpd.GeoDataFrame"],
        np.ndarray,
        list,
        str,
        float,
        int,
        bool,
    ],
    use_json: bool = True,
) -> Any:
    """
    Cast python types to jsonifyable message.

    Parameters
    ----------
    schema
    use_json
        Added for rollout reasons. Use old json formatting until every backend and the frontend is updated.

    Returns
    -------
    message
        jsonifyable message.
    """
    use_json = not bool(os.environ.get("FORCE_SERIALIZATION", not use_json))
    if gpd is not None:
        if isinstance(schema, (gpd.GeoDataFrame, gpd.GeoSeries)):
            if use_json:
                return serialize_geopandas_json(schema)
            return NotImplemented
    if isinstance(schema, pl.DataFrame):
        if use_json:
            return serialize_polars_json(schema)
        return NotImplemented
    if isinstance(schema, (pd.DataFrame, pd.Series)):
        if use_json:
            return serialize_pandas_json(schema)
        return serialize_pandas_parquet(schema)
    if isinstance(schema, np.ndarray):
        if use_json:
            # no NaN in array
            if all(isinstance(x, (float, int)) for x in schema.flatten()):
                return json.loads(json.dumps(schema.tolist()).replace("NaN", "null"))
            else:
                return schema.tolist()
        return serialize_numpy(schema)
    # recurse and convert to jsonifyable types
    if isinstance(schema, Mapping):
        return {k: python_types_to_message(v) for k, v in schema.items()}
    if isinstance(schema, Collection) and not isinstance(schema, str):
        return [python_types_to_message(k) for k in schema]
    return schema


def message_to_python_types(
    message: Union[
        dict,
        list,
        str,
        float,
        int,
        bool,
    ]
) -> Union[
    dict,
    pd.Series,
    pd.DataFrame,
    pl.DataFrame,
    Type["gpd.GeoSeries"],
    Type["gpd.GeoDataFrame"],
    np.ndarray,
    list,
    str,
    float,
    int,
    bool,
]:
    """
    Cast json result to python types by searching for objects

    Parameters
    ----------
    message
        parsed json message

    Returns
    -------
    schema
        schema with original python types.
    """

    if isinstance(message, dict):
        # input is a custom serialized type
        if "nuclei" in message and len(message) == 2:
            t = message["nuclei"]["type"]
            if t == DF_PARQUET:
                warnings.warn(
                    "`DataFrame.parquet` is deprecated and will soon be replaced by `PANDAS_DataFrame.parquet`.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return deserialize_pandas_parquet(message)
            elif t == PANDAS_DF_PARQUET:
                return deserialize_pandas_parquet(message)
            elif t == DF or t == PDS:
                warnings.warn(
                    "`DataFrame` and `Series` is deprecated and will soon be replaced by `PANDAS_DataFrame` or `PANDAS_Series`.",
                    DeprecationWarning,
                    stacklevel=2,
                )
                return deserialize_pandas_json(message)
            elif t == PANDAS_DF or t == PANDAS_PDS:
                return deserialize_pandas_json(message)
            elif t == GDF or t == GPDS:
                return deserialize_geopandas_json(message)
            elif t == POLARS_DF:
                return deserialize_polars_json(message)
            elif t == NUMPY:
                return deserialize_numpy(message)

        # input is a "normal" dict
        return {k: message_to_python_types(v) for k, v in message.items()}

    # input is a list, tuple or set
    if isinstance(message, Collection) and not isinstance(message, str):
        return [message_to_python_types(k) for k in message]

    # return original input if it is anything else
    return message


def to_json(data: str) -> dict:
    return json.loads(data.replace("'", '"'))


def validify_url(url: str) -> str:
    """
    Remove leading part of URL
    """
    # Using 'ttps' because look-behind requires fixed-width pattern
    return re.sub(r"(?<!http:|ttps:)//", r"/", url)


def create_routing_table() -> Dict[str, Tuple[str, str]]:
    baseurl = "https://crux-nuclei.com"
    r = requests.get(baseurl + "/g/svc/api")
    if r.status_code != 200:  # pragma: no cover
        raise IOError("Could not connect to nuclei for routing tables")

    routes = dict()

    for app in r.json():
        # json schema example:
        # app: "gef-model"
        # url: "/api/gef-model/"
        # host_env: "GEF_MODEL_SERVICE_HOST"
        # port_env: "GEF_MODEL_SERVICE_PORT"

        app_name = app["app"]
        url = app["url"]
        host = os.environ.get(app["host_env"])
        port = os.environ.get(app["port_env"])
        routes[app_name] = (baseurl + url, f"http://{host}:{port}{url}")
    return routes


def token_time_valid(jwt: str) -> bool:
    """
    Check if the jwt is not expired.
    """
    payload = jwt.split(".")[1]
    # Here we decode the base64 encoded middle part of the JWT token. The length of the
    # payload needs to be a multi-fold of 4, so we pad it with "=" dummies if necessary.
    try:
        content: Dict = json.loads(
            base64.urlsafe_b64decode(
                payload + "=" * divmod(len(payload), 4)[1]
            ).decode()
        )
    except (binascii.Error, UnicodeDecodeError):
        return False

    # expiry time in seconds since unix epoch
    exp = content.get("exp")
    if exp is not None:
        return int(exp) > time.time()
    return True



from nuclei._version import __version__
from nuclei.api_zoo import call_endpoint, get_applications, get_endpoints

__all__ = [
    "__version__",
    "call_endpoint",
    "get_applications",
    "get_endpoints",
]



import base64
import getpass
import os
import time
from functools import lru_cache
from typing import Any, List, Optional, Union

import requests
from IPython.display import Image

from nuclei import utils

ROUTING = {
    app: urls
    for app, urls in utils.create_routing_table().items()
    if app not in ["redis", "gef-model-legacy"]
}
INTERNAL = int(os.environ.get("PRODUCTION", 0))
ACCESS_TOKEN: Optional[str] = None
MAX_RECURSION_COUNT = 10


def get_url(app: str) -> str:
    """
    Get API's url

    Parameters
    ----------
    app : str

    Returns
    -------
    url : str
    """
    return ROUTING[app][INTERNAL]


def get_applications() -> List[str]:
    """
    Get available API's in the Nuclei landscape.

    Returns
    -------
    out : list[str]
        Names of the API's
    """
    return list(ROUTING.keys())


@lru_cache(16)
def get_app_specification(app: str) -> dict:
    return requests.get(get_url(app) + "openapi.json").json()


def get_endpoints(app: str) -> List[str]:
    """
    Get available endpoints of single API.

    Parameters
    ----------
    app : str
        Name of the API.

    Returns
    -------
    out : list[str]
        Endpoint urls.
    """
    return list(get_app_specification(app)["paths"].keys())


def get_endpoint_type(app: str, endpoint: str) -> str:
    """
    Parameters
    ----------
    app
        name of the app
    endpoint
        url of the endpoint.

    Returns
    -------
    "get" | "post"
    """
    return list(get_app_specification(app)["paths"][endpoint].keys())[0]


def resolve_ref(spec: dict, ref: str) -> dict:
    """
    Parameters
    ----------
    spec
        full openapi spec of the app
    ref
        reference to look up

    Returns
    -------
    Part of the schema at the ref
    """
    if not ref.startswith("#/"):
        raise ValueError(f"Invalid OpenAPI ref {ref}")  # pragma: no cover
    path = ref[2:].split("/")
    for p in path:
        if p not in spec:
            raise ValueError(f"Unable to resolve OpenAPI ref {ref}")  # pragma: no cover
        spec = spec[p]
    return spec


def flatten_schema_parameters(spec: dict, schema: dict) -> List[str]:
    """
    Used to extract request body parameters from an openapi spec.
    Handles $ref, allOf, anyOf, and oneOf.

    Parameters
    ----------
    spec
        full openapi spec of the app, used to resolve $ref references
    schema
        schema part of the spec to flatten

    Returns
    -------
    List of properties of the root object of the schema
    """
    if schema.get("type") == "object":
        return list(schema.get("properties", {}).keys())
    if "$ref" in schema:
        return flatten_schema_parameters(spec, resolve_ref(spec, schema["$ref"]))
    for key in ["allOf", "anyOf", "oneOf"]:
        if key in schema:
            return [p for s in schema[key] for p in flatten_schema_parameters(spec, s)]
    if "type" in schema:  # pragma: no cover
        return []
    raise ValueError("Unable to parse OpenAPI spec")  # pragma: no cover


@lru_cache(16)
def get_endpoint_parameters(app: str, endpoint: str, method: str = None) -> List[str]:
    """
    Parameters
    ----------
    app
        name of the app
    endpoint
        url of the endpoint
    method
        http method of the endpoint.

    Returns
    -------
    List of parameter names
    """
    if method is None:
        method = get_endpoint_type(app, endpoint)
    else:
        method = method.lower()
    app_spec = get_app_specification(app)
    spec = app_spec["paths"][endpoint][method]
    if method == "get" or "parameters" in spec:
        return [p["name"] for p in spec.get("parameters", [])]
    else:
        request_body = spec.get("requestBody", {})
        if "$ref" in request_body:
            request_body = resolve_ref(app_spec, request_body["$ref"])
        return flatten_schema_parameters(
            app_spec,
            request_body.get("content", {})
            .get("application/json", {})
            .get("schema", {}),
        )


def call_endpoint(
    app: str,
    endpoint: str,
    schema: Union[dict, str],
    return_response: bool = False,
    get_result: bool = True,
    **kwargs: int,
) -> Any:
    """
    Calls an API in the nuclei landscape.

    Parameters
    ----------
    app : str
        Name of the API. call `get_applications` to obtain a list with all applications.
    endpoint : str
        Name of the API's endpoint. call `get_endpoints` to obtain a list with all applications for a given API.
    schema : dict
        The parameter schema for the API. Take a look at the API documentation.
    return_response : bool
        Return the requests response instead of the parsed data object.
    get_result : bool
        Automatically call the /get-result/ endpoint if we receive a task in PENDING status.

    Returns
    -------
    out : dict
        Parsed API's json response.
    """
    validate_or_refresh_token(ACCESS_TOKEN)

    t = get_endpoint_type(app, endpoint)

    if isinstance(schema, str):
        schema = utils.to_json(schema)

    if (
        get_result
        and "delay" not in schema
        and "delay" in get_endpoint_parameters(app, endpoint, t)
    ):
        schema = dict(delay=True, **schema)

    if t == "get":
        r = requests.get(
            utils.validify_url(get_url(app) + endpoint),
            params=utils.python_types_to_message(schema),
            headers={"Authorization": f"Bearer {ACCESS_TOKEN}"},
        )
    else:
        r = requests.post(
            utils.validify_url(get_url(app) + endpoint),
            json=utils.python_types_to_message(schema),
            headers={"Authorization": f"Bearer {ACCESS_TOKEN}"},
        )

    content_type = r.headers["Content-Type"]
    if (
        get_result
        and content_type == "application/json"
        and "/get-result/" in get_app_specification(app)["paths"]
    ):
        result = r.json()
        if (
            isinstance(result, dict)
            and "id" in result
            and result.get("state") == "PENDING"
        ):
            recursion_count = kwargs.get("recursion_count", 0)
            get_result_schema = {"id": result["id"]}
            timeout = min(10, 2 ** recursion_count)
            if "timeout" in get_endpoint_parameters(app, "/get-result/"):
                time.sleep(0.5)
                get_result_schema["timeout"] = timeout - 0.5
            else:
                time.sleep(timeout)
            return call_endpoint(
                app,
                "/get-result/",
                get_result_schema,
                return_response,
                get_result=recursion_count < MAX_RECURSION_COUNT,
                recursion_count=recursion_count + 1,
            )

    if return_response:
        return r

    if content_type == "image/png;base64":
        return Image(base64.b64decode(r.text))
    elif content_type == "image/png":
        return Image(r.content)
    elif content_type == "application/json":
        result = r.json()
        if "msg" in result:
            if result["msg"] == "Token has expired":
                authenticate()
                return call_endpoint(app, endpoint, schema)

        try:
            python_types = utils.message_to_python_types(result)
        except Exception as e:
            print("Nuclei parsing error", e, result)
            return result
        return python_types
    elif content_type.startswith("text/"):
        return r.text
    return r.content


def authenticate() -> None:
    global ACCESS_TOKEN
    if INTERNAL:
        r = requests.get(
            f"http://{os.environ.get('AUTHENTICATON_SERVER_SERVICE_HOST')}"
            f":{os.environ.get('AUTHENTICATON_SERVER_SERVICE_PORT')}"
            "/get-token"
        )
    else:
        if "NUCLEI_USER" in os.environ and "NUCLEI_PASSWORD" in os.environ:
            username = os.environ["NUCLEI_USER"]
            password = os.environ["NUCLEI_PASSWORD"]
        else:
            username = input("Authentication is needed. What is your nuclei username?")
            password = getpass.getpass("What is your password?")
        r = requests.post(
            "https://crux-nuclei.com/login",
            json={"username": username, "password": password},
        )

    if r.status_code == 401:
        raise Exception("Unauthorized - incorrect username and/or password")
    elif r.status_code != 200:
        raise Exception(f"Cannot authorize (status {r.status_code})")

    ACCESS_TOKEN = r.json()["access_token"]


def validate_or_refresh_token(token: Any) -> int:
    """
    Validates the type and time-stamp of the JWT-string and refreshes the token if necessary.
    Returns an int for testing purposes (1 = token refreshed, 2 = token was valid).
    """
    # type is ok, due to short circuiting
    if (
        (not isinstance(token, str))
        or ("." not in token)
        or (not utils.token_time_valid(token))
    ):
        authenticate()
        return 1
    else:
        return 2



from setuptools import setup, find_packages

exec(open("nuclei/_version.py").read())

setup(
    name="cems-nuclei",
    version=__version__,
    description="Python wrapper around NUCLEI's functionality.",
    author="Ritchie Vink",
    author_email="info@cemsbv.nl",
    url="https://github.com/cemsbv/nuclei",
    license="mit",
    packages=find_packages(),
    install_requires=[
        "requests>=2.21.0",
        "numpy>=1.16.1",
        "pandas>=0.24.1",
        "ipython>=7.3.0",
        "pyarrow>=0.16.0",
        "polars>=0.7.0",
    ],
    extras_require={"geo": ["geopandas>=0.8.0"]},
    python_requires=">=3.7",
)
